# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from functools import update_wrapper, wraps

from django.db import models as m
from django.utils.encoding import force_bytes
import six

from . import WeChatApp


def appmethod(func_or_name):
    """可以把方法注册到WeChatApp上的语法糖

        class SomeModel(WeChatModel):
            @classmethod
            @appmethod
            def awesome_method(cls, app, *args, **kwargs):
                # ...

            @classmethod
            @appmethod("another_method")
            def method(cls, app, *args, **kwargs):
                # ...

        res = SomeModel.awesome_method(app, *args, **kwargs)
        res = app.awesome_method(*args, **kwargs)
        res = app.another_method(*args, **kwargs)
    """
    def decorator(func):
        @wraps(func)
        def decorated_func(*args, **kwargs):
            return func(*args, **kwargs)
        decorated_func._appmethod = name

        return decorated_func

    if callable(func_or_name):
        name = func_or_name.__name__
        return decorator(func_or_name)
    else:
        name = func_or_name
        return decorator


class WeChatModelMetaClass(m.base.ModelBase):
    def __new__(meta, name, bases, attrs):
        # python2.7 __str__必须返回bytestring
        if six.PY2 and "__str__" in attrs:
            __str__ = attrs.pop("__str__")
            attrs["__str__"] = update_wrapper(
                lambda self: force_bytes(__str__(self)), __str__)

        cls = super(WeChatModelMetaClass, meta).__new__(
            meta, name, bases, attrs)
        return cls

    def __init__(cls, name, bases, attrs):
        super(WeChatModelMetaClass, cls).__init__(name, bases, attrs)
        for attr in attrs:
            # 将modelmethod转换为appmethod
            value = getattr(cls, attr)
            if getattr(value, "_appmethod", False):
                method = value

                def wrapped_func(self, *args, **kwargs):
                    return method(self, *args, **kwargs)

                method_name = method._appmethod
                appmethod = wraps(method)(wrapped_func)
                appmethod.__name__ = str(method_name)
                appmethod.__qualname__ = "{0}.{1}".format(
                    "WeChatApp", method_name)
                setattr(WeChatApp, method_name, appmethod)


class WeChatModel(six.with_metaclass(WeChatModelMetaClass, m.Model)):
    class Meta(object):
        abstract = True



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from functools import wraps
import logging

from django.conf.urls import url
from django.http import response
from django.views.decorators.csrf import csrf_exempt
from django.views.decorators.http import require_http_methods
import requests
from wechatpy.constants import WeChatErrorCode
from wechatpy.exceptions import WeChatClientException

from ..models import WeChatApp, WeChatInfo
from ..utils.web import auto_response


def patch_request(request, appname=None, cls=None, **kwargs):
    cls = cls or WeChatInfo
    appname = appname or request.wechat.appname
    wechat = cls(_appname=appname, _request=request)
    for key, value in kwargs.items():
        setattr(wechat, "_" + key, value)
    request.wechat = wechat
    return request


class WeChatSite(object):
    name = "wechat_django"

    def wechat_view(self, view, methods=None):
        """通过wechat_view装饰的view
        request变为``wechat_django.requests.WeChatHttpRequest``实例
        request中将带有``wechat_django.models.WeChatInfo``类型的wechat属性
        """
        methods = methods or ("GET",)

        @wraps(view)
        def decorated_view(request, appname, *args, **kwargs):
            # 只允许queryset内的appname访问本站点
            try:
                app = self.app_queryset.get_by_name(appname)
            except WeChatApp.DoesNotExist:
                return response.HttpResponseNotFound()

            request = patch_request(request, appname, _app=app)
            resp = view(request, *args, **kwargs)
            return auto_response(resp)

        rv = csrf_exempt(decorated_view)
        rv = require_http_methods(methods)(rv)

        return rv

    def get_url(self):
        from ..handler import handler

        route = lambda pattern: r"^(?P<appname>[-_a-zA-Z\d]+)/" + pattern

        return [
            url(
                route(r"materials/(?P<media_id>[-_a-zA-Z\d]+)$"),
                self.wechat_view(self.material_proxy),
                name="material_proxy"
            ),
            url(
                route(r"$"),
                self.wechat_view(handler, ("GET", "POST")),
                name="handler"
            )
        ]

    @property
    def urls(self):
        return self.get_url(), "wechat_django", self.name

    @property
    def app_queryset(self):
        """本站点能查询到的所有app
        :rtype: wechat_django.models.app.WeChatAppQuerySet
        """
        return WeChatApp.objects.get_queryset()

    def material_proxy(self, request, media_id):
        """代理下载微信的素材"""
        app = request.wechat.app
        try:
            resp = app.client.material.get(media_id)
        except WeChatClientException as e:
            if e.errcode == WeChatErrorCode.INVALID_MEDIA_ID:
                return response.HttpResponseNotFound()
            self.get_logger(request.wechat.appname).warning(
                "an exception occurred when download material",
                exc_info=True)
            return response.HttpResponseServerError()
        if not isinstance(resp, requests.Response):
            # 暂时只处理image和voice
            return response.HttpResponseNotFound()

        rv = response.FileResponse(resp.content)
        for k, v in resp.headers.items():
            if k.lower().startswith("content-"):
                rv[k] = v
        return rv

    def get_logger(self, appname):
        return logging.getLogger("wechat.site.{0}".format(appname))


default_site = WeChatSite()
"""默认微信站点,适用于一般状况"""



# -*- coding: utf-8 -*-
from __future__ import unicode_literals


def next_chunk(iterator, count=100):
    rv = []
    for item in iterator:
        rv.append(item)
        if len(rv) >= count:
            yield rv
            rv = []
    if rv:
        yield rv



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import django
from django import forms
from django.contrib import admin
from django.template.defaultfilters import truncatechars
from django.urls import reverse
from django.utils.translation import ugettext_lazy as _

from ..models import MsgLogFlag, WeChatApp
from ..models.permission import get_user_permissions
from .base import has_wechat_permission
from .utils import list_property


class WeChatAppForm(forms.ModelForm):
    log_message = forms.BooleanField(
        label=_("log messages"), initial=False, required=False)

    accesstoken_url = forms.URLField(
        label=_("accesstoken url"), required=False,
        help_text=_("获取accesstoken的url,不填直接从微信取"))
    oauth_url = forms.URLField(
        label=_("oauth url"), required=False,
        help_text=_("授权重定向的url,用于第三方网页授权换取code,默认直接微信授权"))

    class Meta(object):
        model = WeChatApp
        fields = "__all__"
        widgets = dict(
            appsecret=forms.PasswordInput(render_value=True),
            token=forms.PasswordInput(render_value=True),
            encoding_aes_key=forms.PasswordInput(render_value=True)
        )

    def __init__(self, *args, **kwargs):
        inst = kwargs.get("instance")
        if inst:
            initial = kwargs.get("initial", {})
            initial["log_message"] = inst.log_message
            initial["accesstoken_url"] = inst.configurations.get("ACCESSTOKEN_URL", "")
            initial["oauth_url"] = inst.configurations.get("OAUTH_URL", "")
            kwargs["initial"] = initial
        return super(WeChatAppForm, self).__init__(*args, **kwargs)

    def clean(self):
        cleaned_data = super(WeChatAppForm, self).clean()
        if cleaned_data.get("log_message"):
            cleaned_data["flags"] = MsgLogFlag.LOG_MESSAGE
        else:
            cleaned_data["flags"] = 0
        return cleaned_data

    def save(self, commit=True):
        self.instance.flags = self.cleaned_data["flags"]
        self.instance.configurations["ACCESSTOKEN_URL"] =\
            self.cleaned_data.get("accesstoken_url", "")
        self.instance.configurations["OAUTH_URL"] =\
            self.cleaned_data.get("oauth_url", "")
        return super(WeChatAppForm, self).save(commit)


@admin.register(WeChatApp)
class WeChatAppAdmin(admin.ModelAdmin):
    change_form_template = "admin/change_form.html"
    change_list_template = "admin/change_list.html"
    delete_confirmation_template = "admin/delete_confirmation.html"
    object_history_template = "admin/object_history.html"

    actions = None
    list_display = (
        "title", "name", "type", "appid", "short_desc", 
        list_property(
            "abilities.interactable",
            boolean=True, short_description=_("interactable")),
        "created_at", "updated_at")
    search_fields = ("title", "name", "appid", "short_desc")

    fields = (
        "title", "name", "appid", "appsecret", "type", "token",
        "encoding_aes_key", "encoding_mode", "desc", "log_message",
        "callback", "accesstoken_url", "oauth_url",
        "created_at", "updated_at"
    )

    def short_desc(self, obj):
        return truncatechars(obj.desc, 35)
    short_desc.short_description = _("description")

    def callback(self, obj):
        return obj and self.request.build_absolute_uri(reverse(
            "wechat_django:handler", kwargs=dict(appname=obj.name)))
    callback.short_description = _("message callback url")

    def get_urls(self):
        urlpatterns = super(WeChatAppAdmin, self).get_urls()
        # django 1.11 替换urlpattern为命名式的
        if django.VERSION[0] < 2:
            for pattern in urlpatterns:
                pattern._regex = pattern._regex.replace(
                    "(.+)", "(?P<object_id>.+)")
        return urlpatterns

    def get_fields(self, request, obj=None):
        fields = list(super(WeChatAppAdmin, self).get_fields(request, obj))
        if not obj:
            fields.remove("callback")
            fields.remove("created_at")
            fields.remove("updated_at")
        if obj and obj.type == WeChatApp.Type.SUBSCRIBEAPP:
            fields.remove("oauth_url")
        return fields

    def get_readonly_fields(self, request, obj=None):
        rv = super(WeChatAppAdmin, self).get_readonly_fields(request, obj)
        if obj:
            rv = rv + ("name", "appid", "type", "callback",
                "created_at", "updated_at")
        return rv

    def get_queryset(self, request):
        self.request = request
        rv = super(WeChatAppAdmin, self).get_queryset(request)

        # 非管理员 过滤拥有的微信号权限
        if not super(WeChatAppAdmin, self).has_change_permission(request):
            allowed_apps = self._get_allowed_apps(request)
            rv = rv.filter(name__in=allowed_apps)

        return rv

    def has_change_permission(self, request, obj=None):
        rv = super(WeChatAppAdmin, self).has_change_permission(request, obj)
        if not rv:
            if obj:
                return has_wechat_permission(request, obj, "manage")
            else:
                return bool(self._get_allowed_apps(request))
        return rv

    def _get_allowed_apps(self, request):
        """有权限的微信号"""
        perms = get_user_permissions(request.user)
        return set(
            appname for appname, permissions in perms.items()
            if "manage" in permissions
        )

    def get_model_perms(self, request):
        return ({} if getattr(request, "app_id", None)
            else super(WeChatAppAdmin, self).get_model_perms(request))

    form = WeChatAppForm



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models as m
from django.utils.module_loading import import_string
from django.utils.translation import ugettext_lazy as _
from jsonfield import JSONField
import six
from wechatpy.crypto import WeChatCrypto

from .. import settings
from ..utils.model import enum2choices
from . import MsgLogFlag


class Abilities(object):
    """微信号能力"""

    def __init__(self, app=None):
        """:type app: wechat_django.WeChatApp"""
        self._app = app

    @property
    def interactable(self):
        """是否可与微信进行消息交互"""
        rv = self._app.token
        if self._app.encoding_mode == WeChatApp.EncodingMode.SAFE:
            rv = rv and self._app.encoding_aes_key
        return bool(rv)

    @property
    def api(self):
        """是否可调用微信api(可换取accesstoken)"""
        return bool(self._app.appsecret)

    @property
    def oauth(self):
        """是否可进行网页授权"""
        return bool(
            self.authed
            and self._app.type == WeChatApp.Type.SERVICEAPP
            and self._app.appsecret)

    @property
    def menus(self):
        """是否可配置菜单"""
        return self._app.type == WeChatApp.Type.SERVICEAPP\
            or self._app.type == WeChatApp.Type.SUBSCRIBEAPP and self.authed
    
    @property
    def template(self):
        """发送模板消息"""
        return self.authed and bool(self._app.type in (
            WeChatApp.Type.SERVICEAPP, WeChatApp.Type.MINIPROGRAM))

    @property
    def authed(self):
        """已认证"""
        return bool(WeChatApp.Flag.UNAUTH ^ (
            self._app.flags & WeChatApp.Flag.UNAUTH))


class WeChatAppQuerySet(m.QuerySet):
    def get_by_name(self, name):
        return self.get(name=name)


class WeChatAppManager(m.Manager.from_queryset(WeChatAppQuerySet)):
    pass


class WeChatApp(m.Model):
    @staticmethod
    def __new__(cls, *args, **kwargs):
        self = super(WeChatApp, cls).__new__(cls)
        self.abilities = Abilities(self)
        return self

    class EncodingMode(object):
        PLAIN = 0
        # BOTH = 1
        SAFE = 2

    class Flag(object):
        UNAUTH = 0x01 # 未认证

    class Type(object):
        SERVICEAPP = 1
        SUBSCRIBEAPP = 2
        MINIPROGRAM = 4

    title = m.CharField(
        _("title"), max_length=16, null=False,
        help_text=_("公众号名称,用于后台辨识公众号"))
    name = m.CharField(
        _("name"), max_length=16, blank=False, null=False, unique=True,
        help_text=_("公众号唯一标识,可采用微信号,设定后不可修改,用于程序辨识"))
    desc = m.TextField(_("description"), default="", blank=True)

    appid = m.CharField(_("AppId"), max_length=32, null=False)
    appsecret = m.CharField(_("AppSecret"), max_length=64, blank=True, null=True)
    type = m.PositiveSmallIntegerField(_("type"), default=Type.SERVICEAPP,
        choices=enum2choices(Type))

    token = m.CharField(max_length=32, null=True, blank=True)
    encoding_aes_key = m.CharField(
        _("EncodingAESKey"), max_length=43, null=True, blank=True)
    encoding_mode = m.PositiveSmallIntegerField(
        _("encoding mode"), default=EncodingMode.PLAIN,
        choices=enum2choices(EncodingMode))

    flags = m.IntegerField(_("flags"), default=0)

    ext_info = JSONField(db_column="ext_info", default={})
    configurations = JSONField(db_column="configurations", default={})

    created_at = m.DateTimeField(_("created at"), auto_now_add=True)
    updated_at = m.DateTimeField(_("updated at"), auto_now=True)

    objects = WeChatAppManager()

    abilities = Abilities()

    class Meta(object):
        verbose_name = _("WeChat app")
        verbose_name_plural = _("WeChat apps")

    @property
    def log_message(self):
        return bool(self.flags & MsgLogFlag.LOG_MESSAGE)

    @property
    def log_reply(self):
        return bool(self.flags & MsgLogFlag.LOG_REPLY)

    @property
    def client(self):
        """:rtype: wechat_django.client.WeChatClient"""
        if not hasattr(self, "_client"):
            # session
            if isinstance(settings.SESSIONSTORAGE, six.text_type):
                settings.SESSIONSTORAGE = import_string(settings.SESSIONSTORAGE)
            if callable(settings.SESSIONSTORAGE):
                session = settings.SESSIONSTORAGE(self)
            else:
                session = settings.SESSIONSTORAGE

            client_factory = import_string(settings.WECHATCLIENTFACTORY)
            self._client = client = client_factory(self)(
                self.appid,
                self.appsecret,
                session=session
            )
            client.appname = self.name
            # API BASE URL
            if self.configurations.get("ACCESSTOKEN_URL"):
                client.ACCESSTOKEN_URL = self.configurations["ACCESSTOKEN_URL"]

        return self._client

    @property
    def oauth(self):
        if not hasattr(self, "_oauth"):
            from ..oauth import WeChatOAuthClient
            client_factory = import_string(settings.OAUTHCLIENTFACTORY)
            self._oauth = client_factory(self)(self.appid, self.appsecret)

            if self.configurations.get("OAUTH_URL"):
                self._oauth.OAUTH_URL = self.configurations["OAUTH_URL"]

        return self._oauth

    @property
    def crypto(self):
        if not hasattr(self, "_crypto"):
            self._crypto = (self.encoding_mode == self.EncodingMode.SAFE
                and self.abilities.interactable
                and WeChatCrypto(
                    self.token,
                    self.encoding_aes_key,
                    self.appid
                )) or None
        return self._crypto

    def __str__(self):
        rv = "{title} ({name})".format(title=self.title, name=self.name)
        if six.PY2:
            rv = rv.encode("utf-8")
        return rv



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.test import override_settings
from wechatpy.client import WeChatClient as _Client

from .base import mock, WeChatTestCase
from .interceptors import wechatapi, wechatapi_accesstoken, wechatapi_error


class AppTestCase(WeChatTestCase):
    def test_getaccesstoken(self):
        """测试accesstoken获取"""
        api = "/cgi-bin/token"
        # 测试获取accesstoken
        with wechatapi_accesstoken(lambda url, req, resp: self.assertTrue(req)):
            token = self.app.client.access_token
            self.assertEqual(token, "ACCESS_TOKEN")
        # 测试获取后不再请求accesstoken
        success = dict(
            errcode=0,
            errmsg=""
        )
        with wechatapi_error(api), wechatapi("/cgi-bin/message/custom/send", success):
            resp = self.app.client.message.send_text("openid", "abc")
            self.assertEqual(resp["errcode"], 0)

    @override_settings(CACHES={
        "default": {
            'BACKEND': 'django.core.cache.backends.dummy.DummyCache',
        }
    })
    def test_custom_accesstoken_url(self):
        """测试设置了ACCESSTOKEN url后 不再向原地址发送请求,转为向新地址发送请求"""
        with mock.patch.object(_Client, "_fetch_access_token"):
            new_url = "new_url"
            self.app.configurations["ACCESSTOKEN_URL"] = new_url
            hasattr(self.app, "_client") and delattr(self.app, "_client")
            self.app.client.access_token
            self.assertEqual(
                _Client._fetch_access_token.call_args[0][0], new_url)
            delattr(self.app, "_client")
            del self.app.configurations["ACCESSTOKEN_URL"]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from ..utils.func import next_chunk
from .base import WeChatTestCase


class UtilFunctoolTestCase(WeChatTestCase):
    def test_next_chunk(self):
        """测试next chunk"""
        data = list(next_chunk(range(50)))
        self.assertEqual(len(data), 1)
        self.assertEqual(set(range(50)), set(data[0]))

        data = list(next_chunk(range(100)))
        self.assertEqual(len(data), 1)
        self.assertEqual(set(range(100)), set(data[0]))

        data = list(next_chunk(range(101), 100))
        self.assertEqual(len(data), 2)
        self.assertEqual(set(range(100)), set(data[0]))
        self.assertEqual(set((100, )), set(data[1]))



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from setuptools import find_packages, setup

import wechat_django

with open("README.md", "rb") as f:
    long_description = f.read().decode("utf-8")

with open("requirements.txt") as f:
    requirements = [l for l in f.read().splitlines() if l]

setup(
    name=wechat_django.__title__,
    version=wechat_django.__version__,
    author=wechat_django.__author__,
    author_email=wechat_django.__author_email__,
    url=wechat_django.__url__,
    packages=find_packages(),
    keywords="WeChat, weixin, wx, micromessenger, 微信, django",
    description=wechat_django.__description__,
    long_description=long_description,
    long_description_content_type="text/markdown",
    install_requires=requirements,
    include_package_data=True,
    tests_require=requirements + ["cryptography>=2.5", "httmock==1.2.6"],
    test_suite="runtests.main",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Environment :: Web Environment",
        "Framework :: Django :: 1.11",
        "Framework :: Django :: 2.0",
        "Framework :: Django :: 2.1",
        "Framework :: Django :: 2.2",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: Chinese (Simplified)",
        "Natural Language :: Chinese (Traditional)",
        "Natural Language :: English",
        "Operating System :: MacOS",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: POSIX",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: Implementation :: CPython",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Internet :: WWW/HTTP :: Dynamic Content",
        "Topic :: Internet :: WWW/HTTP :: Site Management",
        "Topic :: Software Development :: Libraries",
        "Topic :: Software Development :: Libraries :: Application Frameworks",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Utilities"
    ],
    extras_require={
        "cryptography": ["cryptography"],
        "pycrypto": ["pycryptodome"],
    }
)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from .base import mock, WeChatTestCase


class UserTestCase(WeChatTestCase):
    def test_sync(self):
        """测试同步用户"""
        pass

    def test_fetch_users(self):
        """测试拉取用户"""
        pass

    def test_upsert_users(self):
        """测试插入或更新用户"""
        pass

    def test_update(self):
        """测试更新用户"""
        pass



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.conf import settings

PATCHADMINSITE = getattr(settings, "WECHAT_PATCHADMINSITE", True)
"""是否将django默认的adminsite替换为wechat_django默认的adminsite, 默认替换"""

SESSIONSTORAGE = getattr(
    settings, "WECHAT_SESSIONSTORAGE", "django.core.cache.cache")

WECHATCLIENTFACTORY = getattr(
    settings, "WECHAT_WECHATCLIENTFACTORY",
    "wechat_django.client.get_client")

OAUTHCLIENTFACTORY = getattr(
    settings, "WECHAT_OAUTHCLIENTFACTORY",
    "wechat_django.oauth.get_client")

MESSAGETIMEOFFSET = getattr(settings, "WECHAT_MESSAGETIMEOFFSET", 180)

MESSAGENOREPEATNONCE = getattr(settings, "WECHAT_MESSAGENOREPEATNONCE", True)
